/*
Copyright (C) 2007 Jean-Michel Richer

intrinsics should be found in :

/usr/lib/gcc/i486-linux-gnu/4.1.2/include/xmmintrin.h

*/

#include <xmmintrin.h>
#include <stdio.h>
#include <stdlib.h>
#include "test_and_func.h"

char *prog_name;
int check_mode = 0;

#define ALIGN2(len) (((len)+1) & ~1)
#define ALIGN16(len) (((len)+15) & ~15)

// #define COUNT 1

#define MAX_SEQ     867
#define SEQ_LENGTH  960
#define MAX_LOOPS  20000
// #define MAX_LOOPS  1

unsigned long int count=0;

void (*func)(TYPE *vA, TYPE *vB, TYPE *vC, int len);


TYPE **seq;

int use_alignment = 0;


void allocate_and_set_memory() {
  int i, j;
  size_t size = (MAX_SEQ<<1)*sizeof(TYPE *);

  if (use_alignment) {
    seq = (TYPE **) _mm_malloc(size+16,16);
  } else {
    seq = (TYPE **) malloc(size);
  }

  printf("sequences allocated at address %x\n", seq);

  for (i = 0; i <MAX_SEQ; i++) {
    if (use_alignment) {
      seq[i] = (TYPE *) _mm_malloc(SEQ_LENGTH*sizeof(TYPE),16);
    } else {
      seq[i] = calloc(SEQ_LENGTH, sizeof(TYPE));
    }
    for (j = 0; j <SEQ_LENGTH; j++) {
      //seq[i][j] = 1 << (rand() % 7);
      seq[i][j] =  ((i % 2) << 1) | (j%2);
    }
  }
  while (i < ((MAX_SEQ<<1)-1)) {
    if (use_alignment) {
      seq[i] = (TYPE *) _mm_malloc(SEQ_LENGTH*sizeof(TYPE),16);
    } else {
      seq[i] = calloc(SEQ_LENGTH, sizeof(TYPE));
    }
    ++i;
  }
}

void checksum() {
}


/**
 * usage
 *
 */
void usage() {
  printf("%s [-t -c]\n", prog_name);
  printf("\twhere n=1,2,..,8");
  printf("\n\t-t display function title");
  printf("\n\t-c check result");
  printf("\n\nexamples :");
  printf("\n\t./%s 1", prog_name);
  printf("\n\t./%s -t 1", prog_name);
  printf("\n\t./%s -c 1", prog_name);
  printf("\n");
  exit(1);
}

#define MAX_FUNCTIONS 8

char *function_desc[] = {
  "non_available",
  "basic_function",
  "unroll_4",
  "unroll_8",
  "unroll_16",
  "unsigned_int",
  "unsigned_int_unroll_4",
  "sse",
  "sse_unroll_4"
};

/**
 * main
 *
 */
int main(int argc, char *argv[]) {
  int i;
  int n;

  srand(1);

  prog_name = argv[0];

  switch(argc) {
  case 2:
    n = atoi(argv[1]);
    break;

  case 3:
    if (strcmp(argv[1],"-c")==0) {
      check_mode = 1;
      n = atoi(argv[2]);
      if ((n<1) || (n>MAX_FUNCTIONS)) usage();
    } else {
      if (strcmp(argv[1],"-t")==0) {
	n = atoi(argv[2]);
	if ((n<1) || (n>MAX_FUNCTIONS)) usage();
	printf("%s", function_desc[n]);
	exit(1);
      } else usage();
    }
    break;

  default: usage();
    break;
  }

  switch(n) {
  case 1:
    printf("%s\n", function_desc[1]);
    func = func1;
    break;

  case 2:
    printf("%s\n", function_desc[2]);
    func = func_unroll4;
    break;

  case 3:
    printf("%s\n", function_desc[3]);
    func = func_unroll8;
    break;

  case 4:
    printf("%s\n", function_desc[4]);
    func = func_unroll16;
    break;

  case 5:
    printf("%s\n", function_desc[5]);
    func = func_uint;
    break;

  case 6:
    printf("%s\n", function_desc[6]);
    func = func_uint_unroll4;
    break;

  case 7:
    printf("%s\n", function_desc[7]);
    use_alignment=1;
    func = func_xmm;
    break;

  case 8:
    printf("%s\n", function_desc[8]);
    use_alignment=1;
    func = func_xmm_unroll4;
    break;


  default:
    usage();
  }

  allocate_and_set_memory();

  if (check_mode) {

      int j = 0;
      int k = MAX_SEQ;
      int l;
      long int total=0;

      while (j < k-1) {
	(*func)(seq[j], seq[j+1], seq[k], SEQ_LENGTH);
	for (l=0; l<SEQ_LENGTH; ++l) total += seq[k][l];
	j+=2;
	++k;
      }
      printf("total found %ld\n", total);
      if (total == 415680L) 
	printf("\nOK\n");
      else
	printf("\nERROR : expected value 415680\n"); 
  

  } else { // check_mode

    for (i=0; i<MAX_LOOPS; ++i) {
      int j = 0;
      int k = MAX_SEQ;

      while (j < k-1) {
	(*func)(seq[j], seq[j+1], seq[k], SEQ_LENGTH);
	j+=2;
	++k;
      }
    }

  } // endif check_mode


  /*
  for (i=0; i<SEQ_LENGTH; ++i) {
    printf("%3d : %3d\n", i, seq[MAX_SEQ][i]);
  }
  */
  return 0;
}

