/* Copyright (C) 2007 Jean-Michel Richer */
#include "test_and_func.h"
#include <xmmintrin.h>

/**
 * basic function : perform and
 */
void func1(TYPE *vA, TYPE *vB, TYPE *vC, int len) {
  int i;

  for (i = 0; i < len; i++) {
    vC[i] = vA[i] & vB[i];
  }
}

/**
 * unroll loop by 4
 *
 */
void func_unroll4(TYPE *vA, TYPE *vB, TYPE *vC, int len) {
  int i;

  for (i = 0; i < len; i+=4) {
    vC[i] = vA[i] & vB[i];
    vC[i+1] = vA[i+1] & vB[i+1];
    vC[i+2] = vA[i+2] & vB[i+2];
    vC[i+3] = vA[i+3] & vB[i+3];
  }
  
  for (; i < len; i++) {
    vC[i] = vA[i] & vB[i];
  }
}


/**
 * unroll loop by 8
 *
 */
void func_unroll8(TYPE *vA, TYPE *vB, TYPE *vC, int len) {
  int i;

  for (i = 0; i < len; i+=8) {
    vC[i] = vA[i] & vB[i];
    vC[i+1] = vA[i+1] & vB[i+1];
    vC[i+2] = vA[i+2] & vB[i+2];
    vC[i+3] = vA[i+3] & vB[i+3];
    vC[i+4] = vA[i+4] & vB[i+4];
    vC[i+5] = vA[i+5] & vB[i+5];
    vC[i+6] = vA[i+6] & vB[i+6];
    vC[i+7] = vA[i+7] & vB[i+7];
  }
  
  for (; i < len; i++) {
    vC[i] = vA[i] & vB[i];
  }
}


/**
 * unroll loop by 16
 *
 */
void func_unroll16(TYPE *vA, TYPE *vB, TYPE *vC, int len) {
  int i;

  for (i = 0; i < len; i+=16) {
    vC[i] = vA[i] & vB[i];
    vC[i+1] = vA[i+1] & vB[i+1];
    vC[i+2] = vA[i+2] & vB[i+2];
    vC[i+3] = vA[i+3] & vB[i+3];
    vC[i+4] = vA[i+4] & vB[i+4];
    vC[i+5] = vA[i+5] & vB[i+5];
    vC[i+6] = vA[i+6] & vB[i+6];
    vC[i+7] = vA[i+7] & vB[i+7];
    vC[i+8] = vA[i+8] & vB[i+8];
    vC[i+9] = vA[i+9] & vB[i+9];
    vC[i+10] = vA[i+10] & vB[i+10];
    vC[i+11] = vA[i+11] & vB[i+11];
    vC[i+12] = vA[i+12] & vB[i+12];
    vC[i+13] = vA[i+13] & vB[i+13];
    vC[i+14] = vA[i+14] & vB[i+14];
    vC[i+15] = vA[i+15] & vB[i+15];
  }
  
  for (; i < len; i++) {
    vC[i] = vA[i] & vB[i];
  }
}


/**
 * treat bytes by 4 using unsigned integers
 *
 */
void func_uint(TYPE *vA, TYPE *vB, TYPE *vC, int len) {
  int i=0;
  int len4 = len >> 2;
  unsigned int *uA = (unsigned int *) vA;
  unsigned int *uB = (unsigned int *) vB;
  unsigned int *uC = (unsigned int *) vC;

  for (i = 0; i<len4; ++i) {
    uC[i] = uA[i] & uB[i];
  }
  i = i << 2;
  while (i < len) {
    vC[i] = vA[i] & vB[i];
    ++i;
  }
}

/**
 * treat bytes by 4 using unsigned integers and unroll4
 *
 */
void func_uint_unroll4(TYPE *vA, TYPE *vB, TYPE *vC, int len) {
  int i=0;
  int len4 = len >> 2;
  unsigned int *uA = (unsigned int *) vA;
  unsigned int *uB = (unsigned int *) vB;
  unsigned int *uC = (unsigned int *) vC;

  for (i = 0; i<len4; i+=4) {
    uC[i] = uA[i] & uB[i];
    uC[i+1] = uA[i+1] & uB[i+1];
    uC[i+2] = uA[i+2] & uB[i+2];
    uC[i+3] = uA[i+3] & uB[i+3];
  }
  i = i << 2;
  while (i < len) {
    vC[i] = vA[i] & vB[i];
    ++i;
  }
}


void func_xmm_c(TYPE *vA, TYPE *vB, TYPE *vC, int len) {
  int i;

  const __m128 *xmm1 = (__m128 *) vA;
  const __m128 *xmm2 = (__m128 *) vB;
   __m128 *xmm3 = (__m128 *) vC;

  len = len >> 4;
  for (i = 0; i < len; i++) {
    *xmm3 = _mm_and_ps(*xmm1,*xmm2);
    ++xmm1;
    ++xmm2;
    ++xmm3;
  }
}
